/**
* @fileOverview Array is one of the core classes to native JavaScript. Array mixes in the Enumerable class
*  which gives Array a lot more functionality. See Enumerable to see all the additional methods on Array.
* @example
*/

/**
* Sets or returns the number of elements in an array.
* @name length
* @type {Number}
*/
/** @ignore */
Array.prototype._$length$_ = "don't use me, used for reflection.";

/**
* Joins two or more arrays and returns the result.
* @example
["Tom", "Dick", "Harry"].concat(["Matt", "Annie", "Sam"]); -> ["Tom", "Dick", "Harry", "Matt", "Annie", "Sam"]
* @name concat
* @param {Array} _args One or more array objects to be joined to an array. Array#unshift(_args1, _args2,.....,_argsX) allowed.
* @return {Array}
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$concat$_ = function(_args){throw "don't call me, used for reflection.";}

/**
* Puts all the elements of an array into a string. The elements are separated by a specified delimiter.
* @example
["Tom", "Dick", "Harry"].join(" - "); -> "Tom - Dick - Harry"
* @name join
* @param {String} [_sep] Specifies the separator to be used. Comma is the default separator.
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$join$_ = function(_sep){throw "don't call me, used for reflection.";}

/**
* Removes and returns the last element of an array.
* @example
["Tom", "Dick", "Harry"].pop(); -> "Harry"
* @name pop
* @return {Object}
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$pop$_ = function(){throw "don't call me, used for reflection.";}

/**
* Adds one or more elements to the end of an array and returns the new length.
* @example
["Tom", "Dick", "Harry"].push("Matt"); -> 4, contents = ["Tom", "Dick", "Harry", "Matt"]
* @name push
* @param {Object} _args One or more items. Array#push(_args1,.....,_argsX) allowed.
* @return {Number}
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$push$_ = function(_args){throw "don't call me, used for reflection.";}

/**
* Reverses the order of the elements in an array.
* @example
["Tom", "Dick", "Harry"].reverse(); -> ["Harry", "Dick", "Tom"]
* @name reverse
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$reverse$_ = function(){throw "don't call me, used for reflection.";}

/**
* Removes and returns the first element of an array.
* @example
["Tom", "Dick", "Harry"].shift(); -> "Tom", contents = ["Dick", "Harry"]
* @name shift
* @return {Object}
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$shift$_ = function(){throw "don't call me, used for reflection.";}

/**
* Returns selected elements from an existing array.
* @example
["Tom", "Dick", "Harry"].slice(1,3); -> ["Dick", "Harry"]
* @name slice
* @param {Number} _start
* @param {Number} [_end]
* @return {Array}
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$slice$_ = function(_start, _end){throw "don't call me, used for reflection.";}

/**
* Sorts the elements of an array.
* @example
["Tom", "Dick", "Harry"].sort(); -> ["Dick", "Harry", "Tom"]
* @name sort
* @param {Function} [_func] The block function of the sort implementation
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$sort$_ = function(_func){throw "don't call me, used for reflection.";}

/**
* Removes and adds new elements to an array.
* @example
<div>["Tom", "Dick", "Harry", "Matt", "Annie", "Sam"].splice(2, 0, "Carson"); -> contents = ["Tom", "Dick", "Carson", "Harry", "Matt", "Annie", "Sam"]</div>
<div>["Tom", "Dick", "Harry", "Matt", "Annie", "Sam"].splice(2, 1, "Carson"); -> contents = ["Tom", "Dick", "Carson", "Matt", "Annie", "Sam"]</div>
* @param {Number} _start Specify where to add/remove elements
* @param {Number} _deleteCount Specify how many elements should be removed. Greater then zero.
* @param {Object} [_args] Specify a new element to add to the array. Array#splice(_start, _deleteCount, _args1,.....,_argsX) allowed.
* @name splice
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$splice$_ = function(_start, _deleteCount, _args){throw "don't call me, used for reflection.";}

/**
* Adds one or more elements to the beginning of an array and returns the new length.
* @example
["Tom", "Dick", "Harry"].unshift("Matt", "Annie", "Sam"); -> 6, contents = ["Tom", "Dick", "Harry", "Matt", "Annie", "Sam"]
* @name unshift
* @param {Object} _args1 The first element to add to the array.
* @param {Object} [_args2] The second element to add to the array. Array#unshift(_args1, _args2,.....,_argsX) allowed.
* @function
* @memberOf Array
*/
/** @ignore */
Array.prototype._$unshift$_ = function(_args1, _args2){throw "don't call me, used for reflection.";}

/**
* Returns whether this object is equal to the specified _object.
* @example
["Tom", "Dick", "Harry"].equals(["Tom", "Dick", "Harry"]); -> true;
* @param {Object} _object
* @return {Boolean}
*/
Array.prototype.equals = function(_object) {

	if(this === _object) {

		return true;
	}
	if(this.constructor != _object.constructor) {

		return false;
	}

	// compare lengths
	if(this.size() != _object.size()){ return false; }

	// compare entries in the indexes
	for(var i=0;i<this.length;i++) {

		// if both element are null pass them.
		if((this[i] == null) && (_object[i] == null)){ continue; }

		if(!this[i].equals) {

			if(this[i] != _object[i]) {

				return false;
			}
		} else {

			if(!this[i].equals(_object[i])) {

				return false;
			}
		}
	}

	return true;
};

/**
* Returns a description of the array
* @return {String}
*/
Array.prototype.toString = function() {

	var me = this;
	var sb = "";
	sb += "[";

	this.each(function(_value, _index) {

		sb += _value.toString();
		if(_index != (me.size() -1)){ sb += ", "; }
	});

	sb += "]";
	return sb;
};

// enumerable methods
/**
* Required by the <code>Enumerable</code> mixin. Calls <code>Array#iterator(_block)</code> passing a value and index to the block.
* @param {Function} _block. Parameters passed to the _block are Object and the Number.
*/
Array.prototype.each = function(_block) {

	var me = this;
	var index = 0;
	try {

		this.iterator(function(value) {

			try {

				_block(value, index++);

			} catch(e) {

				if(e != Enumerable.$continue){ throw e; }
			}
		});

	} catch(e) {

		if(e != Enumerable.$break){ throw e; }
	}
}

/**
* Iterates through the array, executing the _block and passing a value from Array[index] as its parameter. Used by <code>Array#each(_block)</code>.
* @param {Function} _block. Parameter passed to the _block is the element object at the index.
*/
Array.prototype.iterator = function(_block) {

	for(var i=0;i<this.length;i++) {

		_block(this[i]);
	}
}

Array.prototype.$break = Enumerable.$break;
Array.prototype.$continue = Enumerable.$continue;
Array.prototype.echo = Enumerable.echo;
Array.prototype.convertToArray = Enumerable.convertToArray;
Array.prototype.all = Enumerable.all;
Array.prototype.any = Enumerable.any;
Array.prototype.collect = Enumerable.collect;
Array.prototype.detect = Enumerable.detect;
Array.prototype.findAll = Enumerable.findAll;
Array.prototype.grep = Enumerable.grep;
Array.prototype.include = Enumerable.include;
Array.prototype.inject = Enumerable.inject;
Array.prototype.invoke = Enumerable.invoke;
Array.prototype.max = Enumerable.max;
Array.prototype.min = Enumerable.min;
Array.prototype.partition = Enumerable.partition;
Array.prototype.pluck = Enumerable.pluck;
Array.prototype.reject = Enumerable.reject;
Array.prototype.sortBy = Enumerable.sortBy;
Array.prototype.zip = Enumerable.zip;
Array.prototype.toArray = Enumerable.toArray;
Array.prototype.entries = Enumerable.entries;
Array.prototype.map = Enumerable.map;
Array.prototype.find = Enumerable.find;
Array.prototype.member = Enumerable.member;
Array.prototype.select = Enumerable.select;

// querying methods
/**
* Returns the element at the specified position in this list.
* @param {Number} _index
* @return {Object}
*/
Array.prototype.get = function(_index) {

	return this[_index];
};

/**
* Returns the index in this array of the first occurence of the specified element, or -1 if the array does not contain this element.
* @example
["Tom", "Dick", "Harry", "Matt", "Annie", "Sam"].indexOf("Annie"); -> 4
* @param {Object} _object
* @return {Number}
*/
Array.prototype.indexOf = function(_object) {

	var me = this;
	var result = -1;
	this.inject(0, function(_memo, _value, _index) {

		if(_object == null) {

			if(_value == null) {

				result = _index;
				me.$break();
			}
		} else if ((_value.equals &&_value.equals(_object)) || _value == _object) {

			result = _index;
			me.$break();
		}

		return _memo;
	});
	return result;
}

/**
* Returns the index in this list of the last occurence of the specified element, or -1 if the list does not contain this element.
* @param {Object} _object
* @return {Number}
*/
Array.prototype.lastIndexOf = function(_object) {

	var result = -1;
	this.inject(0, function(_memo, _value, _index) {

		if(_object == null) {

			if(_value == null) {

				result = _index;
			}
		} else if (_value.equals(_object)) {

			result = _index;
		}

		return _memo;
	});
	return result;
}

/**
* Returns whether the specified object is contained in the array.
* @param {Object} _object Parameter could be null. If there is a null element in the array, true will be returned.
* @return {Boolean}
*/
Array.prototype.contains = function(_object) {

	var me = this;
	if(_object == null) {

		// check for null
		var result = false;
		this.each(function(_value) {

			if(_value == null) {

				result = true;
				me.$break();
			}
		});
		return result;

	} else {

		return this.include(_object);
	}
};

/**
* Returns whether the elements in the specified array exist in the array.
* @param {array}
* @return {Boolean}
*/
Array.prototype.containsAll = function(_array) {

	var me = this;
	var result = true;
	_array.each(function(_value) {

		if(!me.contains(_value)) {

			result = false;
			me.$break();
		}
	});
	return result;
}

/**
* Returns the top item off of this stack without removing it.
* @return {Object}
*/
Array.prototype.peek = function() {

	var size = this.size();

	if(size == 0) {
		return null;
	}

	return this.get(size - 1);
}

/**
* Returns the one-based position of the distance from the top that the specified object exists on this stack, where the top-most element is considered to be at distance 1.
* @param {Object} _object
* @return {Number}
*/
Array.prototype.search = function(_object) {

	var index = this.lastIndexOf(_object);
	if(index > -1) {

		return this.size() - index;
	}
	return -1;
}

/**
* Returns the size of the array.
* @return {Number}
*/
Array.prototype.size = function() {

	return this.length;
};

/**
* Returns true|false if the array is empty
* @return {Boolean}
*/
Array.prototype.isEmpty = function() {

	return this.size() == 0;
};

/**
* Returns a view of the portion of this list between fromIndex, inclusive, and toIndex, exclusive.
* @param {Number} _fromIndex
* @param {Number} _toIndex
* @reutrn {List}
*/
Array.prototype.subList = function(_fromIndex, _toIndex) {

	var sublist = [];

	this.inject(0, function(_memo, _value, _index) {

		if(_index >= _fromIndex && _index <= _toIndex) {

			sublist.add(_value);
		}

		return _memo;
	});

	return sublist;
};

// mutating methods

/**
* Replaces the element at the specified position in this list with the specified element. Returns the element previously at the specified position.
* @param {Number} _index
* @param {Object} _object
* @return {Object}
*/
Array.prototype.set = function(_index, _object) {

	this[_index] = _object;
};

/**
* Inserts the specified element at the specified position in this list.
* @param {Number} _index
* @param {Object} _object
*/
Array.prototype.addAt = function(_index, _object) {

	if(_index < 0){return false;}
	// adding at _index, removing 0 items
	this.splice(_index, 0, _object);
	return true;
};

/**
* Inserts all of the elements in the specified array into this list at the specified position.
* @param {Number} _index
* @param {Array} _array
*/
Array.prototype.addAllAt = function(_index, _array) {

	var me = this;
	var modified = false;
	_array.each(function(_object, _itemIndex) {

		me.addAt((_index + _itemIndex), _object);
		modified = true;
	});
	return modified;
};

/**
* Appends the specified element to the end of this list. Returns whether the element was added to the list.
* @param {Object} _object
* @return {Boolean}
*/
Array.prototype.add = function(_object) {

	if(_object && !_object.equals){ _object.equals = function(_object){ return this == _object; }; }

	return this.addAt(this.size(), _object);
};

/**
* Adds all the elements in the specified collecion to the end of the array. Returns whether the current array was modified.
* @param {Array} _array
* @return {Boolean}
*/
Array.prototype.addAll = function(_array) {

	var me = this;
	var modified = false;
	_array.each(function(_value) {

		if(me.add(_value)) {

			modified = true;
		}
	});
	return modified;
};

/**
* Removes and returns the element at the specified position in this list.
* @param {Number} _index
* @return {Object}
*/
Array.prototype.removeAt = function(_index) {

	if(_index < 0){ return false; }

	var items = this.splice(_index, 1);

	if(items.length == 0){ return false; }

	return true;
};

/**
* Removes the specified object from the array. Return whether the object was removed.
* @param {Object} _object
* @return {Boolean}
*/
Array.prototype.remove = function(_object) {

	var index = this.indexOf(_object);
	if(index > -1) {

		return this.removeAt(index);
	}
	return false;
};

/**
* Removes all the elements in the specified array from the current array. Returns whether the current array was modified.
* @param {Array} _array
* @return {Boolean}
*/
Array.prototype.removeAll = function(_array) {

	var me = this;
	var modified = false;
	_array.each(function(_value) {

		if(me.remove(_value)) {

			modified = true;
		}
	});
	return modified;
};

/**
* Removes all elements in the current array if they are not in the specified array. Returns whether the current array was modified.
* @param {Object} _array
* @return {Boolean}
*/
Array.prototype.retainAll = function(_array) {

	var me = this;
	var modified = false;

	var tmp = [];
	this.each(function(_value) {

		if(!_array.contains(_value)) {

			tmp[tmp.length] = _value;
			modified = true;
		}
	});

	if(modified) {

		for(var i=0;i<tmp.length;i++) {

			this.remove(tmp[i]);
		}
	}

	return modified;
};

/**
* Clears the array
*/
Array.prototype.clear = function() {

	this.length = 0;
};

// sorting methods

Array.prototype.getDefaultComparator = function(_reversal) {

	return function(_objectA, _objectB) {

		return Array.defaultCompare(_objectA, _objectB, _reversal);
	};
}

Array.prototype.getNumberComparator = function(_reversal) {

	return function(_objectA, _objectB) {

		return Array.numberCompare(_objectA, _objectB, _reversal);
	};
}

Array.prototype.getStringComparator = function(_reversal, _insensitiveComparator) {

	return function(_objectA, _objectB) {

		return Array.stringCompare(_objectA, _objectB, _reversal, _insensitiveComparator);
	};
}

Array.prototype.getDateComparator = function(_reversal) {

	return function(_objectA, _objectB) {

		return Array.dateCompare(_objectA, _objectB, _reversal);
	};
}

Array.prototype.getCurrencyComparator = function(_reversal) {

	return function(_objectA, _objectB) {

		return Array.currencyCompare(_objectA, _objectB, _reversal);
	};
}

Array.defaultCompare = function(_objectA, _objectB, _reversal) {

	var a = _objectA;
	var b = _objectB;

	var comp;

	if(a < b){ comp = -1; }
	if(a == b){ comp = 0; }
	if(a > b){ comp = 1; }

	if(_reversal) {

		return -1 * comp;
	}

	return comp;
};

Array.numberCompare = function(_objectA, _objectB, _reversal) {

	var a = _objectA;
	var b = _objectB;

	if(Number.isNaN(a)){ a = 0; }
	if(Number.isNaN(b)){ b = 0; }

	return Array.defaultCompare(a, b, _reversal);
};

Array.stringCompare = function(_objectA, _objectB, _reversal, _insensitiveComparator) {

	var a = _objectA;
	var b = _objectB;

	if(_insensitiveComparator) {

		a = a.toLowerCase();
		b = b.toLowerCase();
	}

	return Array.defaultCompare(a, b, _reversal);
};

Array.dateCompare = function(_objectA, _objectB, _reversal) {

	var _10digit = function(_date) {

		return _date.substr(6,4)+_date.substr(0,2)+_date.substr(3,2);
	};

	var _less10digit = function(_date) {

		var yr = _date.substr(6,2);
		if(Number.parseInt(yr) < 50) { yr = "20"+yr; } else { yr = "19"+yr; }
		return yr+_date.substr(0,2)+_date.substr(3,2);
	};

	var a = _objectA;
	var b = _objectB;

	// format month/day/year
	// y2k notes: two digit years less than 50 are treated as 20XX, greater than 50 are treated as 19XX

	if(a.length == 10) {

		a = _10digit(a);

	} else {

		a = _less10digit(a);
	}

	if(b.length == 10) {

		b = _10digit(b);

	} else {

		b = _less10digit(b);
	}

	return Array.defaultCompare(a, b, _reversal);
};

Array.currencyCompare = function(_objectA, _objectB, _reversal) {

	var a = _objectA;
	var b = _objectB;

	a = a.replace(/[^0-9.]/g, '');
	b = b.replace(/[^0-9.]/g, '');

	a = Number.parseFloat(a);
	b = Number.parseFloat(b);

	return Array.defaultCompare(a, b, _reversal);		
};
